from pallets_sphinx_themes import get_version

# Project --------------------------------------------------------------

project = "magql"
copyright = "2019 Moebius Solutions"
release, version = get_version("magql")

# General --------------------------------------------------------------

master_doc = "index"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "pallets_sphinx_themes",
    "sphinxcontrib.log_cabinet",
    "sphinx_issues",
]
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}
issues_github_path = "autoinvent/magql"

# HTML -----------------------------------------------------------------

html_theme = "flask"
html_sidebars = {
    "index": ["searchbox.html"],
    "**": ["localtoc.html", "relations.html", "searchbox.html"],
}
singlehtml_sidebars = {"index": ["localtoc.html"]}
html_title = f"{project} Documentation ({version})"
html_show_sourcelink = False
html_domain_indices = False



from functools import singledispatch

from inflection import camelize
from sqlalchemy import Boolean
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import JSON
from sqlalchemy import Numeric
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy import Time
from sqlalchemy import Unicode
from sqlalchemy import UnicodeText
from sqlalchemy import VARCHAR
from sqlalchemy_utils import ChoiceType
from sqlalchemy_utils import EmailType
from sqlalchemy_utils import JSONType
from sqlalchemy_utils import PhoneNumberType
from sqlalchemy_utils import URLType

from .definitions import MagqlBoolean
from .definitions import MagqlEnumType
from .definitions import MagqlFloat
from .definitions import MagqlInt
from .definitions import MagqlNonNull
from .definitions import MagqlString
from .filter import BooleanFilter
from .filter import DateFilter
from .filter import EnumFilter
from .filter import FloatFilter
from .filter import IntFilter
from .filter import StringFilter


@singledispatch
def _get_magql_type(type_, column):
    """
    Returns the corrsponding GraphQL type to the given SQLA column type
    :param type: The type of the SQLA column
    :return: The corresponding GraphQL type
    """
    raise TypeError(
        "No type registered for column type"
        f" {type_.__class__.__name__!r} from {column!r}."
    )


def get_magql_type(col):
    return _get_magql_type(col.type, col)


def is_required(col):
    """
    Checks whether a scalar SQLAlchemy column is required or not
    :param col: SQLAlchemy column
    :return: Whether or not the column is required
    """
    return not col.nullable and not col.default and not col.primary_key


def get_magql_required_type(col):
    type_ = get_magql_type(col)
    if is_required(col):
        return MagqlNonNull(type_)
    else:
        return type_


@_get_magql_type.register(JSON)
@_get_magql_type.register(JSONType)
@_get_magql_type.register(DateTime)
@_get_magql_type.register(Text)
@_get_magql_type.register(Date)
@_get_magql_type.register(UnicodeText)
@_get_magql_type.register(Unicode)
@_get_magql_type.register(URLType)
@_get_magql_type.register(PhoneNumberType)
@_get_magql_type.register(EmailType)
@_get_magql_type.register(Time)
@_get_magql_type.register(String)
@_get_magql_type.register(VARCHAR)
def _get_string_type(type, column):
    # if "image" in column.info:
    #     return MagqlFile()
    return MagqlString()


@_get_magql_type.register(Boolean)
def _get_boolean_type(type, column):
    return MagqlBoolean()


@_get_magql_type.register(Integer)
def _get_integer_type(type, column):
    return MagqlInt(MagqlInt.parse_value_accepts_string)


@_get_magql_type.register(Float)
@_get_magql_type.register(Numeric)
def _get_float_type(type, column):
    return MagqlFloat(MagqlFloat.parse_value_accepts_string)


@_get_magql_type.register(ChoiceType)
def _get_choice_type(type_, column):
    # name = camelize(column.table.name) + camelize(column.name) + "EnumType"
    # enums = dict((key, value) for key, value in type.choices)
    # rm = GraphQLEnumType(name, enums)
    # return rm
    name = camelize(column.table.name) + camelize(column.name) + "EnumType"
    enums = {key: key for key, value in type_.choices}
    return MagqlEnumType(name, enums)


def get_magql_filter_type(type_, base_type):
    return _get_magql_filter_type(type_.type, base_type)


@singledispatch
def _get_magql_filter_type(column, base_type):
    """
    Returns the filter based on the given type
    :param type: The type of the SQLAlchemy column
    :return:  The Autogenerated GraphQL filter object
    """
    raise TypeError(
        f"No filter registered for column type {column.__class__.__name__!r}."
    )


@_get_magql_filter_type.register(JSON)
@_get_magql_filter_type.register(JSONType)
@_get_magql_filter_type.register(DateTime)
@_get_magql_filter_type.register(Text)
@_get_magql_filter_type.register(Date)
@_get_magql_filter_type.register(UnicodeText)
@_get_magql_filter_type.register(Unicode)
@_get_magql_filter_type.register(URLType)
@_get_magql_filter_type.register(PhoneNumberType)
@_get_magql_filter_type.register(EmailType)
@_get_magql_filter_type.register(Time)
@_get_magql_filter_type.register(String)
@_get_magql_filter_type.register(VARCHAR)
def _get_string_filter(type, base_type):
    return StringFilter


@_get_magql_filter_type.register(Date)
@_get_magql_filter_type.register(DateTime)
def _get_date_filter(type, base_type):
    return DateFilter


@_get_magql_filter_type.register(Integer)
def _get_integer_filter(type, base_type):
    return IntFilter


@_get_magql_filter_type.register(Float)
@_get_magql_filter_type.register(Numeric)
def _get_float_filter(type, base_type):
    return FloatFilter


@_get_magql_filter_type.register(Boolean)
def _get_boolean_filter(type, base_type):
    return BooleanFilter


@_get_magql_filter_type.register(ChoiceType)
def _get_choice_filter(type_, base_type):
    return EnumFilter(base_type)



from datetime import date

import pytest
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import Date
from sqlalchemy import Enum
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker

from magql.definitions import MagqlField
from magql.manager import MagqlTableManager
from magql.manager import MagqlTableManagerCollection

base = declarative_base()
Session = sessionmaker()


@pytest.fixture
def house():
    return House(name="House 1")


@pytest.fixture
def car():
    return Car(name="Car 1")


@pytest.fixture
def hometown():
    return Hometown(id="San Diego", country="United States")


@pytest.fixture
def wealth():
    return Wealth(id=1.5, currency="USD", abbreviation="M")


@pytest.fixture
def person(car, house, hometown, status, wealth):
    return Person(
        name="Person 1",
        car=car,
        house=house,
        hometown=hometown,
        status=status,
        wealth=wealth,
    )


@pytest.fixture
def status():
    return Status(id=True, date_of_birth=date.today())


@pytest.fixture
def single_query():
    return """
query test {
  person(id: 1) {
    result {
      name
    }
  }
  car(id: 2) {
    result {
      name
    }
  }
  house(id: 3) {
    result {
      name
    }
  }
}
"""


@pytest.fixture
def many_query():
    return """
query test {
  people {
    result {
      name
    }
  }
  cars {
    result {
      name
    }
  }
  houses {
    result {
      name
    }
  }
}
"""


@pytest.fixture
def session(car, house, person):
    """Create and configure a new app instance for each test."""
    # create the app with common test config
    engine = create_engine("sqlite://")
    base.metadata.create_all(engine)
    Session.configure(bind=engine)
    session = Session()

    session.add(car)
    session.add(house)
    session.add(person)
    session.commit()
    return session


class DummyInfo:  # noqa: E501
    def __init__(self, session):
        self.context = session


@pytest.fixture
def info(session):
    return DummyInfo(session)


@pytest.fixture
def po_box_manager():
    base_manager = MagqlTableManager(POBox.__table__, "POBox")
    base_manager.single_query_name = "POBox"
    base_manager.many_query_name = "POBoxes"
    base_manager.magql_types["POBox"].fields["customField"] = MagqlField(
        "Int", resolve=lambda x: 1
    )
    return base_manager


@pytest.fixture
def manager_collection(po_box_manager):
    table = {}
    managers = {"pobox": po_box_manager}
    for table_name, _table in base.metadata.tables.items():
        if table_name == "BadClass" or table_name == "BadRelClass":
            continue
        table[table_name] = _table
    return MagqlTableManagerCollection(table, managers=managers)


class Person(base):
    __tablename__ = "Person"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    age = Column(Integer)

    house_id = Column(ForeignKey("House.id"), nullable=False)
    house = relationship("House", back_populates="inhabitants")

    car_id = Column(ForeignKey("Car.id"))
    car = relationship("Car", back_populates="drivers")

    hometown_id = Column(ForeignKey("Hometown.id"))
    hometown = relationship("Hometown", back_populates="notable_people")

    status_id = Column(ForeignKey("Status.id"))
    status = relationship("Status", back_populates="person")

    wealth_id = Column(ForeignKey("Wealth.id"))
    wealth = relationship("Wealth", back_populates="person_of_note")

    po_box_id = Column(ForeignKey("pobox.id"))
    po_box = relationship("POBox", back_populates="owner")


class POBox(base):
    __tablename__ = "pobox"
    id = Column(Integer, primary_key=True)
    number = Column(Integer)

    owner = relationship(
        "Person", cascade="all, delete-orphan", back_populates="po_box"
    )


class House(base):
    __tablename__ = "House"

    id = Column(Integer, primary_key=True)
    name = Column(String)

    inhabitants = relationship(
        "Person", cascade="all, delete-orphan", back_populates="house"
    )


class Car(base):
    __tablename__ = "Car"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    mpg = Column(Float)
    top_speed = Column(Numeric)

    drivers = relationship("Person", back_populates="car")


class Hometown(base):
    __tablename__ = "Hometown"

    id = Column(String, primary_key=True, unique=True, nullable=False)
    country = Column(String)

    notable_people = relationship(
        "Person", cascade="all, delete-orphan", back_populates="hometown"
    )


class Status(base):
    __tablename__ = "Status"

    id = Column(Boolean, primary_key=True, nullable=False)
    date_of_birth = Column(Date, nullable=False)
    date_of_death = Column(Date, nullable=True)

    person = relationship(
        "Person", cascade="all, delete-orphan", back_populates="status"
    )


class Wealth(base):
    __tablename__ = "Wealth"

    id = Column(Float, primary_key=True, nullable=False)
    currency = Column(Enum("USD", "EUR", "GBP", name="denomination"), nullable=False)
    abbreviation = Column(Enum("M", "K", "B", "T", name="money_abbreviation"))

    person_of_note = relationship(
        "Person", cascade="all, delete-orphan", back_populates="wealth"
    )


class BadClass(base):
    """Test class for bad relationship keys"""

    __tablename__ = "BadClass"

    id = Column(Date, primary_key=True, nullable=False)

    bad_rel = relationship(
        "BadRelClass", cascade="all, delete-orphan", back_populates="bad_class"
    )


class BadRelClass(base):
    """Relationship class for bad relationship key"""

    __tablename__ = "BadRelClass"

    id = Column(String, primary_key=True, nullable=False)

    bad_class_id = Column(ForeignKey("BadClass.id"))
    bad_class = relationship("BadClass", back_populates="bad_rel")



from functools import singledispatch

from inflection import underscore
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import Text
from sqlalchemy import Time
from sqlalchemy import Unicode
from sqlalchemy import UnicodeText
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.types import Boolean
from sqlalchemy.types import Float
from sqlalchemy.types import Integer
from sqlalchemy.types import Numeric
from sqlalchemy.types import String
from sqlalchemy.types import VARCHAR
from sqlalchemy_utils import EmailType
from sqlalchemy_utils import get_mapper
from sqlalchemy_utils import JSONType
from sqlalchemy_utils import PhoneNumberType
from sqlalchemy_utils import URLType
from sqlalchemy_utils.types import ChoiceType

from .definitions import MagqlEnumType
from .definitions import MagqlInputField
from .definitions import MagqlInputObjectType

StringFilter = MagqlInputObjectType(
    "StringFilter",
    {
        "operator": MagqlInputField(
            MagqlEnumType(
                "StringOperator",
                {
                    "INCLUDES": "INCLUDES",
                    "EQUALS": "EQUALS",
                    "EXISTS": "EXISTS",
                    "DOESNOTEXIST": "DOESNOTEXIST",
                },
            )
        ),
        "value": MagqlInputField("String"),
    },
)


DateFilter = MagqlInputObjectType(
    "DateFilter",
    {
        "operator": MagqlInputField(
            MagqlEnumType(
                "DateOperator", {"BEFORE": "BEFORE", "ON": "ON", "AFTER": "AFTER"}
            )
        ),
        "value": MagqlInputField("String"),
    },
)


IntFilter = MagqlInputObjectType(
    "IntFilter",
    {
        "operator": MagqlInputField(
            MagqlEnumType(
                "IntOperator",
                {
                    "lt": "lt",
                    "lte": "lte",
                    "eq": "eq",
                    "neq": "neq",
                    "gt": "gt",
                    "gte": "gte",
                },
            )
        ),
        "value": MagqlInputField("Int"),
    },
)

FloatFilter = MagqlInputObjectType(
    "FloatFilter",
    {
        "operator": MagqlInputField(
            MagqlEnumType(
                "FloatOperator",
                {
                    "lt": "lt",
                    "lte": "lte",
                    "eq": "eq",
                    "neq": "neq",
                    "gt": "gt",
                    "gte": "gte",
                },
            )
        ),
        "value": MagqlInputField("Float"),
    },
)

RelFilter = MagqlInputObjectType(
    "RelFilter",
    {
        "operator": MagqlInputField(
            MagqlEnumType("RelOperator", {"INCLUDES": "INCLUDES"})
        ),
        "value": MagqlInputField("Int"),
    },
)


BooleanFilter = MagqlInputObjectType(
    "BooleanFilter",
    {
        "operator": MagqlInputField(
            MagqlEnumType(
                "BooleanOperator", {"EQUALS": "EQUALS", "NOTEQUALS": "NOTEQUALS"}
            )
        ),
        "value": MagqlInputField("Boolean"),
    },
)

EnumOperator = MagqlEnumType("EnumOperator", {"INCLUDES": "INCLUDES"})


def EnumFilter(base_type):
    name = base_type.name + "Filter"

    input_ = {
        "operator": MagqlInputField(EnumOperator),
        "value": MagqlInputField(base_type),
    }
    return MagqlInputObjectType(name, input_)


@singledispatch
def get_filter_comparator(type):
    raise TypeError(f"No comparator registered for {type.__class__.__name__!r}.")


@get_filter_comparator.register(RelationshipProperty)
def _get_relationship_comparator(rel):
    direction = rel.direction.name
    if "TOONE" in direction:

        def condition(filter_value, filter_operator, field):
            if filter_operator == "INCLUDES":
                return field == filter_value

        return condition
    elif "TOMANY" in direction:

        def condition(filter_value, filter_operator, field):
            if filter_operator == "INCLUDES":
                return field.any(field.contains(filter_value))

        return condition


@get_filter_comparator.register(DateTime)
@get_filter_comparator.register(Date)
def _get_date_comparator(_):
    def condition(filter_value, filter_operator, field):
        if filter_operator == "BEFORE":
            return field < filter_value
        elif filter_operator == "ON":
            return field == filter_value
        elif filter_operator == "After":
            return field > filter_value

    return condition


@get_filter_comparator.register(JSONType)
@get_filter_comparator.register(Text)
@get_filter_comparator.register(UnicodeText)
@get_filter_comparator.register(Unicode)
@get_filter_comparator.register(URLType)
@get_filter_comparator.register(PhoneNumberType)
@get_filter_comparator.register(EmailType)
@get_filter_comparator.register(Time)
@get_filter_comparator.register(String)
@get_filter_comparator.register(VARCHAR)
def _get_string_comparator(_):
    def condition(filter_value, filter_operator, field):
        if filter_operator == "INCLUDES":
            return field.like(f"%{filter_value}%")
        elif filter_operator == "EQUALS":
            return field == filter_value
        elif filter_operator == "EXISTS":
            return field.like("%")
        elif filter_operator == "DOESNOTEXIST":
            return field.is_(None)

    return condition


@get_filter_comparator.register(Float)
@get_filter_comparator.register(Numeric)
@get_filter_comparator.register(Integer)
def _get_number_comparator(_):
    def condition(filter_value, filter_operator, field):
        if filter_operator == "lt":
            return field < filter_value
        elif filter_operator == "lte":
            return field <= filter_value
        elif filter_operator == "eq":
            return field == filter_value
        elif filter_operator == "neq":
            return field != filter_value
        elif filter_operator == "gt":
            return field > filter_value
        elif filter_operator == "gte":
            return field >= filter_value

    return condition


@get_filter_comparator.register(Boolean)
def _get_boolean_comparator(_):
    def condition(filter_value, filter_operator, field):
        if filter_operator == "EQUALS":
            return field == filter_value
        elif filter_operator == "NOTEQUALS":
            return field != filter_value

    return condition


@get_filter_comparator.register(ChoiceType)
def _get_choice_comparator(_):
    def condition(filter_value, filter_operator, field):
        if filter_operator == "INCLUDES":
            return field == filter_value

    return condition


def generate_filters(table, info, *args, **kwargs):
    sqla_filters = []
    if "filter" in kwargs and kwargs["filter"] is not None:
        mapper = get_mapper(table)
        gql_filters = kwargs["filter"]
        for filter_name, gql_filter in gql_filters.items():
            gql_filter_value = gql_filter["value"]
            filter_name = underscore(filter_name)
            if filter_name in table.c:
                filter_type = table.c[filter_name].type
            elif filter_name in mapper.relationships:
                rel = mapper.relationships[filter_name]
                rel_mapper = get_mapper(rel.target)
                gql_filter_value = (
                    info.context.query(rel_mapper.class_)
                    .filter_by(id=gql_filter_value)
                    .one()
                )
                filter_type = rel
            else:
                raise KeyError(filter_name)

            sql_filter = get_filter_comparator(filter_type)(
                gql_filter_value,
                gql_filter["operator"],
                getattr(mapper.class_, filter_name),
            )
            sqla_filters.append(sql_filter)
    return sqla_filters



from decimal import Decimal

import pytest
from tests.conftest import base
from tests.conftest import Car
from tests.conftest import House
from tests.conftest import Person

from magql.resolver_factory import CreateResolver
from magql.resolver_factory import DeleteResolver
from magql.resolver_factory import SingleResolver
from magql.resolver_factory import UpdateResolver


def compare(output, test_input):
    for key, value in test_input.items():
        output_value = getattr(output, key)
        if isinstance(output_value, list):
            for instance in output_value:
                if isinstance(instance, base):
                    assert instance.id in value
        elif isinstance(output_value, base):
            assert output_value.id == value
        else:
            assert output_value == value


@pytest.mark.parametrize(
    "input_data",
    [
        (House, {"name": "House 2", "inhabitants": [1]}),
        (
            Car,
            {
                "name": "Car 2",
                "drivers": [1],
                "mpg": 25.6,
                "top_speed": Decimal("105.333"),
            },
        ),
        (Person, {"name": "Person 2", "age": 30, "car": 1, "house": 1}),
    ],
)
def test_create_resolver(input_data, info, session):
    test_class = input_data[0]
    test_input = input_data[1]
    resolve = CreateResolver(test_class.__table__)

    output = resolve(None, info, input=test_input)

    compare(output, test_input)


@pytest.mark.parametrize(
    "input_data",
    [
        (House, 1, {"name": "House 2", "inhabitants": [1]}),
        (
            Car,
            1,
            {
                "name": "Car 2",
                "drivers": [1],
                "mpg": 33.8,
                "top_speed": Decimal("94.825"),
            },
        ),
        (Person, 1, {"name": "Person 2", "age": 30, "car": 1, "house": 1}),
    ],
)
def test_update_resolver(input_data, info, session):
    test_class = input_data[0]
    test_id = input_data[1]
    test_input = input_data[2]
    resolve = UpdateResolver(test_class.__table__)

    output = resolve(None, info, id=test_id, input=test_input)

    compare(output, test_input)


@pytest.mark.parametrize("input_data", [(House, 1), (Car, 1), (Person, 1)])
def test_delete_resolvers(input_data, info, session):
    test_class = input_data[0]
    test_id = input_data[1]
    resolve = DeleteResolver(test_class.__table__)

    resolve(None, info, id=test_id)

    del_inst = session.query(test_class).filter_by(id=test_id).one_or_none()

    assert del_inst is None


@pytest.mark.parametrize("model", [House, Car, Person])
@pytest.mark.parametrize("model_id", [1, 2])
def test_single_resolvers(model, model_id, session, info):
    resolver = SingleResolver(model.__table__)
    resolved_value = resolver(None, info, id=model_id)
    queried_value = session.query(model).filter_by(id=model_id).one_or_none()
    if queried_value is not None:
        assert queried_value == resolved_value
    else:
        assert resolved_value["result"] is None



"""Utils
This module defines the framework-agnostic logic for handling the multipart
request spec
"""
from graphql import GraphQLScalarType

GraphQLFile = GraphQLScalarType(
    name="File",
    description="The `File` scalar type represents"
    " a file input as specified by"
    " https://github.com/jaydenseric"
    "/graphql-multipart-request-spec"
    "and as implemented in "
    "https://github.com/lmcgartland/"
    "graphene-file-upload/blob/master/"
    "graphene_file_upload/utils.py",
    serialize=lambda x: x,
    parse_value=lambda x: x,
    parse_literal=lambda x: x,
)


def place_files_in_operations(operations, files_map, files):
    """Replaces None placeholders in operations with file objects in the files
    dictionary, by following the files_map logic as specified within the 'map'
    request parameter in the multipart request spec"""
    path_to_key_iter = (
        (value.split("."), key)
        for (key, values) in files_map.items()
        for value in values
    )
    # Since add_files_to_operations returns a new dict/list, first define
    # output to be operations itself
    output = operations
    for path, key in path_to_key_iter:
        file_obj = files[key]
        output = add_file_to_operations(output, file_obj, path)
    return output


def add_file_to_operations(operations, file_obj, path):
    """Handles the recursive algorithm for adding a file to the operations
    object"""
    if not path:
        if operations is not None:
            raise ValueError("Path in map does not lead to a null value")
        return file_obj
    if isinstance(operations, dict):
        key = path[0]
        sub_dict = add_file_to_operations(operations[key], file_obj, path[1:])
        return new_merged_dict(operations, {key: sub_dict})
    if isinstance(operations, list):
        index = int(path[0])
        sub_item = add_file_to_operations(operations[index], file_obj, path[1:])
        return new_list_with_replaced_item(operations, index, sub_item)
    raise TypeError("Operations must be a dict or a list of dicts")


def new_merged_dict(*dicts):
    """Merges dictionaries into a new dictionary. Necessary for python2 and
    python34 since neither have PEP448 implemented."""
    # Necessary for python2 support
    output = {}
    for d in dicts:
        output.update(d)
    return output


def new_list_with_replaced_item(input_list, index, new_value):
    """Creates new list with replaced item at specified index"""
    output = [i for i in input_list]
    output[index] = new_value
    return output



import pytest
from inflection import camelize
from inflection import pluralize
from tests.conftest import base
from tests.conftest import Car
from tests.conftest import Hometown
from tests.conftest import House
from tests.conftest import Person
from tests.conftest import Status
from tests.conftest import Wealth

from magql.definitions import MagqlArgument
from magql.definitions import MagqlBoolean
from magql.definitions import MagqlEnumType
from magql.definitions import MagqlField
from magql.definitions import MagqlFloat
from magql.definitions import MagqlInputObjectType
from magql.definitions import MagqlInt
from magql.definitions import MagqlList
from magql.definitions import MagqlNonNull
from magql.definitions import MagqlObjectType
from magql.definitions import MagqlString
from magql.manager import MagqlTableManager
from magql.manager import MagqlTableManagerCollection
from magql.resolver_factory import CreateResolver
from magql.resolver_factory import DeleteResolver
from magql.resolver_factory import ManyResolver
from magql.resolver_factory import SingleResolver
from magql.resolver_factory import UpdateResolver


def check_magql_types(manager):
    magql_name = manager.magql_name
    assert isinstance(manager.magql_types[magql_name], MagqlObjectType)
    assert isinstance(manager.magql_types[magql_name + "Input"], MagqlInputObjectType)
    assert isinstance(
        manager.magql_types[magql_name + "InputRequired"], MagqlInputObjectType
    )
    assert isinstance(manager.magql_types[magql_name + "Filter"], MagqlInputObjectType)
    assert isinstance(manager.magql_types[magql_name + "Sort"], MagqlEnumType)


def bad_manager_collection():
    table = {}
    for table_name, _table in base.metadata.tables.items():
        if table_name == "BadClass" or table_name == "BadRelClass":
            table[table_name] = _table
    return MagqlTableManagerCollection(table)


@pytest.mark.parametrize("model", [House, Car, Person])
def test_default_table_manager(model):
    table = model.__table__
    table_name = model.__tablename__
    manager = MagqlTableManager(table)
    assert model is manager.table_class
    assert table is manager.table
    assert camelize(table_name) == manager.table_name
    assert table_name == manager.table_name
    check_magql_types(manager)


@pytest.mark.parametrize("model", [House, Car, Person])
def test_magql_name_override(model):
    table = model.__table__
    magql_name = model.__tablename__ + "Override"
    manager = MagqlTableManager(table, magql_name)
    assert magql_name == manager.magql_name
    check_magql_types(manager)


class CreateResolverOverride(CreateResolver):
    def resolve(self, parent, info, *args, **kwargs):
        pass


class DeleteResolverOverride(CreateResolver):
    def resolve(self, parent, info, *args, **kwargs):
        pass


class UpdateResolverOverride(CreateResolver):
    def resolve(self, parent, info, *args, **kwargs):
        pass


class ManyResolverOverride(CreateResolver):
    def resolve(self, parent, info, *args, **kwargs):
        pass


class SingleResolverOverride(CreateResolver):
    def resolve(self, parent, info, *args, **kwargs):
        pass


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_names(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    assert manager.single_query_name == camelize(table.name, False)
    assert manager.many_query_name == camelize(pluralize(table.name), False)
    assert manager.create_mutation_name == "create" + manager.magql_name
    assert manager.update_mutation_name == "update" + manager.magql_name
    assert manager.delete_mutation_name == "delete" + manager.magql_name


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field(model):
    table = model.__table__
    manager = MagqlTableManager(table)
    assert isinstance(manager.create, MagqlField)
    assert isinstance(manager.update, MagqlField)
    assert isinstance(manager.delete, MagqlField)
    assert isinstance(manager.single, MagqlField)
    assert isinstance(manager.many, MagqlField)


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field_name(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    assert manager.single.type_name == manager.magql_name + "Payload"
    assert manager.many.type_name == manager.magql_name + "ListPayload"
    assert manager.create.type_name == manager.magql_name + "Payload"
    assert manager.update.type_name == manager.magql_name + "Payload"
    assert manager.delete.type_name == manager.magql_name + "Payload"


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field_single_resolver_arguments(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    single = manager.single

    assert "id" in single.args
    assert isinstance(single.args["id"], MagqlArgument)
    assert isinstance(single.args["id"].type_, MagqlNonNull)
    assert single.args["id"].type_.type_ == "Int"


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field_many_resolver_arguments(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    many = manager.many

    assert "filter" in many.args
    assert "sort" in many.args

    assert isinstance(many.args["filter"], MagqlArgument)
    assert many.args["filter"].type_ == manager.magql_name + "Filter"

    assert isinstance(many.args["sort"], MagqlArgument)
    assert isinstance(many.args["sort"].type_, MagqlList)
    assert isinstance(many.args["sort"].type_.type_, MagqlNonNull)
    assert many.args["sort"].type_.type_.type_ == manager.magql_name + "Sort"


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field_create_resolver_arguments(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    create = manager.create

    assert "input" in create.args
    assert isinstance(create.args["input"], MagqlArgument)
    assert isinstance(create.args["input"].type_, MagqlNonNull)
    assert create.args["input"].type_.type_ == manager.magql_name + "InputRequired"


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field_update_resolver_arguments(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    update = manager.update

    assert "input" in update.args
    assert isinstance(update.args["input"], MagqlArgument)
    assert isinstance(update.args["input"].type_, MagqlNonNull)
    assert update.args["input"].type_.type_ == manager.magql_name + "Input"

    assert "id" in update.args
    assert isinstance(update.args["id"], MagqlArgument)
    assert isinstance(update.args["id"].type_, MagqlNonNull)
    assert update.args["id"].type_.type_ == "Int"


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field_delete_resolver_arguments(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    delete = manager.delete

    assert "id" in delete.args
    assert isinstance(delete.args["id"], MagqlArgument)
    assert isinstance(delete.args["id"].type_, MagqlNonNull)
    assert delete.args["id"].type_.type_ == "Int"


@pytest.mark.parametrize("model", [House, Car, Person])
def test_generated_field_resolvers(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    assert isinstance(manager.single.resolve, SingleResolver)
    assert isinstance(manager.many.resolve, ManyResolver)
    assert isinstance(manager.create.resolve, CreateResolver)
    assert isinstance(manager.update.resolve, UpdateResolver)
    assert isinstance(manager.delete.resolve, DeleteResolver)


@pytest.mark.parametrize("model", [House, Car, Person])
def test_return_types(model):
    table = model.__table__
    manager = MagqlTableManager(table)

    assert manager.single_query_name == camelize(table.name, False)
    assert manager.many_query_name == camelize(pluralize(table.name), False)
    assert manager.create_mutation_name == "create" + manager.magql_name
    assert manager.update_mutation_name == "update" + manager.magql_name
    assert manager.delete_mutation_name == "delete" + manager.magql_name


@pytest.mark.parametrize("model", [House, Car, Person])
def test_resolver_overrides(model):
    table = model.__table__
    manager = MagqlTableManager(
        table,
        None,
        CreateResolverOverride(table),
        UpdateResolverOverride(table),
        DeleteResolverOverride(table),
        SingleResolverOverride(table),
        ManyResolverOverride(table),
    )
    assert isinstance(manager.create.resolve, CreateResolverOverride)
    assert isinstance(manager.update.resolve, UpdateResolverOverride)
    assert isinstance(manager.delete.resolve, DeleteResolverOverride)
    assert isinstance(manager.single.resolve, SingleResolverOverride)
    assert isinstance(manager.many.resolve, ManyResolverOverride)


@pytest.mark.parametrize("model", [House, Car, Hometown, Status, Wealth, Person])
def test_add_rels(model, manager_collection):
    input_required_types = [MagqlInt, MagqlString, MagqlBoolean, MagqlFloat]
    input_required_strings = ["String", "Int", "Boolean", "Float"]
    for v in (
        manager_collection.manager_map[model.__tablename__]
        .magql_types[f"{model.__tablename__}InputRequired"]
        .fields.values()
    ):
        if isinstance(v.type_name, MagqlNonNull) or isinstance(v.type_name, MagqlList):
            field_type = v.type_name.type_
        else:
            field_type = v.type_name
        if type(field_type) in input_required_types:
            check_type = input_required_types[
                input_required_types.index(type(field_type))
            ]
        else:
            check_type = input_required_strings[
                input_required_strings.index(field_type)
            ]
        if isinstance(field_type, str):
            assert field_type == check_type
        else:
            assert isinstance(field_type, check_type)


def test_magql_name_rel_override(manager_collection):
    assert (
        manager_collection.manager_map["Person"]
        .magql_types["Person"]
        .fields["poBox"]
        .type_name
        == "POBox"
    )


def test_custom_fields(manager_collection):
    assert (
        manager_collection.manager_map["pobox"]
        .magql_types["POBox"]
        .fields["customField"]
    )


def test_bad_rels():
    with pytest.raises(KeyError):
        bad_manager_collection()



from sqlalchemy_utils import get_mapper


class SortNotFoundError(Exception):
    def __init__(self, field_name, direction):
        super().__init__(field_name, direction)
        self.field_name = field_name
        self.direction = direction

    def __str__(self):
        return (
            f"Sort not found for"
            f" (field: {self.field_name}, direction: {self.direction})"
        )


def generate_sorts(table, info, *args, **kwargs):
    sqla_sorts = []
    if "sort" in kwargs and kwargs["sort"] is not None:
        class_ = get_mapper(table).class_
        gql_sorts = kwargs["sort"]
        for sort in gql_sorts:
            field_name, direction = sort[0].rsplit("_", 1)
            field = getattr(class_, field_name)
            if direction == "asc":
                sort = field.asc()
            elif direction == "desc":
                sort = field.desc()
            else:
                raise SortNotFoundError(field_name, direction)
            sqla_sorts.append(sort)
    return sqla_sorts



import logging

from inflection import camelize
from inflection import pluralize
from sqlalchemy_utils import get_mapper

from magql.definitions import js_camelize
from magql.definitions import MagqlArgument
from magql.definitions import MagqlEnumType
from magql.definitions import MagqlField
from magql.definitions import MagqlInputField
from magql.definitions import MagqlInputObjectType
from magql.definitions import MagqlInt
from magql.definitions import MagqlList
from magql.definitions import MagqlNonNull
from magql.definitions import MagqlObjectType
from magql.definitions import MagqlUnionType
from magql.filter import RelFilter
from magql.resolver_factory import CamelResolver
from magql.resolver_factory import CheckDeleteResolver
from magql.resolver_factory import CountResolver
from magql.resolver_factory import CreateResolver
from magql.resolver_factory import DeleteResolver
from magql.resolver_factory import EnumResolver
from magql.resolver_factory import ManyResolver
from magql.resolver_factory import Resolver
from magql.resolver_factory import ResultResolver
from magql.resolver_factory import SingleResolver
from magql.resolver_factory import SQLAlchemyTableUnionResolver
from magql.resolver_factory import UpdateResolver
from magql.type import get_magql_filter_type
from magql.type import get_magql_required_type
from magql.type import get_magql_type


def is_rel_required(rel):
    calc_keys = rel._calculated_foreign_keys
    fk = rel._user_defined_foreign_keys.union(calc_keys).pop()
    return not fk.nullable


# TODO: refactor ManagerCollection so it seamlessly integrates regular
# and table managers
class MagqlTableManagerCollection:
    """
    The MagqlTableManagerCollection creates a grouping of related
    managers from the tables that are passed in, if a corresponding
    manager is not already created.
    """

    def __init__(
        self,
        tables,
        managers=None,
        create_resolver=CreateResolver,
        update_resolver=UpdateResolver,
        delete_resolver=DeleteResolver,
        single_resolver=SingleResolver,
        many_resolver=ManyResolver,
    ):
        """
        Creates the managers needed to manange the Magql schema,
        if they have not already been created.
        :param tables: A list of tables to create managers for
        :param managers: A mapping of tables to pre-existing managers
        :param create_resolver: The class to use as the create resolver
        :param update_resolver: The class to use as the update resolver
        :param delete_resolver: The class to use as the delete resolver
        :param single_resolver: The class to use as the single resolver
        :param many_resolver: The class to use as the many resolver
        """
        self.create_resolver = create_resolver
        self.update_resolver = update_resolver
        self.delete_resolver = delete_resolver
        self.single_resolver = single_resolver
        self.many_resolver = many_resolver

        self.manager_map = {}
        for table_name, table in tables.items():
            if managers and table_name in managers:
                manager = managers[table_name]
            else:
                manager = self.generate_manager(table)

            # skip tables that do not have a manager
            if manager:
                manager.to_magql()

            self.manager_map[table_name] = manager

        for _table_name, manager in self.manager_map.items():
            if manager:
                manager.add_rels(self.manager_map)

        self.magql_name_to_table = {}
        self.generate_check_delete()
        self.generate_pagination()

    def generate_check_delete(self):
        check_delete_manager = MagqlManager("checkDelete")

        self.magql_names = []
        for _table_name, manager in self.manager_map.items():
            if manager:
                self.magql_names.append(manager.magql_name)

        for _table_name, manager in self.manager_map.items():
            if isinstance(manager, MagqlTableManager) and manager:
                self.magql_name_to_table[manager.magql_name] = manager.table

        check_delete_manager.magql_types["SQLAlchemyTableUnion"] = MagqlUnionType(
            "SQLAlchemyTableUnion",
            self.magql_names,
            SQLAlchemyTableUnionResolver(self.magql_name_to_table),
        )

        check_delete_manager.query.fields["checkDelete"] = MagqlField(
            MagqlList("SQLAlchemyTableUnion"),
            {
                "tableName": MagqlArgument("String"),
                "id": MagqlArgument(MagqlNonNull("ID")),
            },
            CheckDeleteResolver(list(self.magql_name_to_table.values())),
        )
        self.manager_map["checkDelete"] = check_delete_manager

    def generate_pagination(self):
        page_manager = MagqlManager("PaginationManager")
        page_manager.magql_types["Page"] = MagqlInputObjectType(
            "Page",
            {
                "current": MagqlInputField(MagqlInt()),
                "per_page": MagqlInputField(MagqlInt()),
            },
        )
        self.manager_map["PaginationManager"] = page_manager

    def generate_manager(self, table):
        try:
            get_mapper(table)
        except ValueError:
            logging.getLogger(__name__).warning(f"No mapper for table {table.name!r}.")
            return
        return MagqlTableManager(
            table,
            create_resolver=self.create_resolver(table),
            update_resolver=self.update_resolver(table),
            delete_resolver=self.delete_resolver(table),
            single_resolver=self.single_resolver(table),
            many_resolver=self.many_resolver(table),
        )


class MagqlManager:
    def __init__(self, magql_name):
        self.query = MagqlObjectType("Query")
        self.mutation = MagqlObjectType("Mutation")
        self.magql_types = {}
        # The check delete union type resolver ( and likely more resolvers)
        # relies on the fact that the magql_name and the base object type
        # share the same name
        self.magql_name = magql_name


class MagqlTableManager(MagqlManager):
    """
    The manager used to manage a single sqlalchemy table
    """

    def __init__(
        self,
        table,
        magql_name=None,
        create_resolver=None,
        update_resolver=None,
        delete_resolver=None,
        single_resolver=None,
        many_resolver=None,
    ):
        """
        The manager for a single sqlalchemy table.
        :param table: The table that is being managed
        :param magql_name: Optional name override for how the table is
        referred to
        :param create_resolver: Optional override for create resolver
        :param update_resolver: Optional override for update resolver
        :param delete_resolver: Optional override for delete resolver
        :param single_resolver: Optional override for single resolver
        :param many_resolver: Optional override for many resolver
        """
        super().__init__(
            magql_name if magql_name is not None else camelize(table.name)
        )  # magql_object_name
        # Throws ValueError if it cannot find a table
        self.table_class = get_mapper(table).class_
        self.table = table
        self.table_name = table.name

        self.create_resolver = (
            create_resolver if create_resolver else CreateResolver(self.table)
        )
        self.update_resolver = (
            update_resolver if update_resolver else UpdateResolver(self.table)
        )
        self.delete_resolver = (
            delete_resolver if delete_resolver else DeleteResolver(self.table)
        )
        self.single_resolver = (
            single_resolver if single_resolver else SingleResolver(self.table)
        )
        self.many_resolver = (
            many_resolver if many_resolver else ManyResolver(self.table)
        )

        self.generate_magql_types()

    @property
    def single_query_name(self):
        if hasattr(self, "_single_query_name_override"):
            if callable(self._single_query_name_override):
                return self._single_query_name_override()
            else:
                return self._single_query_name_override
        return js_camelize(self.table.name)

    @single_query_name.setter
    def single_query_name(self, value):
        """
        Overrides the name of the single query to a custom value
        :param value: The name to change the single query to
        """
        self._single_query_name_override = value

    @property
    def many_query_name(self):
        if hasattr(self, "_many_query_name_override"):
            if callable(self._many_query_name_override):
                return self._many_query_name_override()
            else:
                return self._many_query_name_override
        return js_camelize(pluralize(self.table.name))

    @many_query_name.setter
    def many_query_name(self, value):
        """
        Overrides the name of the many query to a custom value
        :param value: The name to change the many query to
        """
        self._many_query_name_override = value

    @property
    def create_mutation_name(self):
        return "create" + self.magql_name

    @property
    def update_mutation_name(self):
        return "update" + self.magql_name

    @property
    def delete_mutation_name(self):
        return "delete" + self.magql_name

    def generate_create_mutation(self):
        # TODO: Move backend auth functions into manager collection
        self.create = MagqlField(
            self.magql_name + "Payload",
            {"input": MagqlArgument(MagqlNonNull(self.magql_name + "InputRequired"))},
            self.create_resolver,
        )

    def generate_update_mutation(self):
        self.update = MagqlField(
            self.magql_name + "Payload",
            {
                "id": MagqlArgument(MagqlNonNull("ID")),
                "input": MagqlArgument(MagqlNonNull(self.magql_name + "Input")),
            },
            self.update_resolver,
        )

    def generate_delete_mutation(self):
        self.delete = MagqlField(
            self.magql_name + "Payload",
            {"id": MagqlArgument(MagqlNonNull("ID"))},
            self.delete_resolver,
        )

    def generate_single_query(self):
        self.single = MagqlField(
            self.magql_name + "Payload",
            {"id": MagqlArgument(MagqlNonNull("ID"))},
            self.single_resolver,
        )

    def generate_many_query(self):
        self.many = MagqlField(
            self.magql_name + "ListPayload",
            {
                "filter": MagqlArgument(self.magql_name + "Filter"),
                "sort": MagqlArgument(
                    MagqlList(MagqlNonNull(self.magql_name + "Sort"))
                ),
                "page": MagqlArgument("Page"),
            },
            self.many_resolver,
        )

    def generate_types(self):
        base = MagqlObjectType(self.magql_name)
        input = MagqlInputObjectType(self.magql_name + "Input")
        input_required = MagqlInputObjectType(self.magql_name + "InputRequired")
        filter_ = MagqlInputObjectType(self.magql_name + "Filter")
        sort = MagqlEnumType(self.magql_name + "Sort")

        for col_name, col in self.table.c.items():
            if col.foreign_keys:
                continue
            field_name = js_camelize(col_name)
            magql_type = get_magql_type(col)
            required_magql_type = get_magql_required_type(col)
            base.fields[field_name] = MagqlField(
                magql_type, None, CamelResolver()
            )  # noqa: E501
            # TODO: Organize better method of having different resolvers
            # for different fields, probably move onto magql_type
            if isinstance(magql_type, MagqlEnumType):
                base.fields[field_name].resolve = EnumResolver()
            if not col.primary_key:
                input.fields[field_name] = MagqlInputField(magql_type)
                input_required.fields[field_name] = MagqlInputField(required_magql_type)
            filter_.fields[field_name] = MagqlInputField(
                get_magql_filter_type(col, magql_type)
            )
            sort.values[field_name + "_asc"] = (col_name + "_asc",)
            sort.values[field_name + "_desc"] = (col_name + "_desc",)

        self.magql_types[self.magql_name] = base

        self.magql_types[self.magql_name + "Input"] = input
        self.magql_types[self.magql_name + "InputRequired"] = input_required
        self.magql_types[self.magql_name + "Filter"] = filter_
        self.magql_types[self.magql_name + "Sort"] = sort

    def generate_magql_types(self):
        self.generate_create_mutation()
        self.generate_update_mutation()
        self.generate_delete_mutation()
        self.generate_single_query()
        self.generate_many_query()

        self.generate_types()

    # Allows fields to be added directly to mutation and query
    def to_magql(self):
        if self.create:
            self.mutation.fields[self.create_mutation_name] = self.create
        if self.update:
            self.mutation.fields[self.update_mutation_name] = self.update
        if self.delete:
            self.mutation.fields[self.delete_mutation_name] = self.delete
        if self.single:
            self.query.fields[self.single_query_name] = self.single
        if self.many:
            self.query.fields[self.many_query_name] = self.many

    # a manager map can be passed in to give information about
    # other managers, such as an overriden name, otherwise a default is used
    def add_rels(self, managers=None):
        try:
            table_mapper = get_mapper(self.table)
        except ValueError:
            logging.getLogger(__name__).warning(
                f"No mapper for table {self.table.name!r}."
            )
            return

        for rel_name, rel in table_mapper.relationships.items():
            rel_table = rel.target

            if rel_table.name in managers:
                rel_manager = managers[rel_table.name]
                if rel_manager is None:
                    continue
            else:
                rel_manager = None
            direction = rel.direction.name
            required = is_rel_required(rel)

            field_name = js_camelize(rel_name)

            # use magql name of rel manager if it exists else use default name
            target_name = (
                rel_manager.magql_name if rel_manager else camelize(rel.target.name)
            )

            base_field = target_name
            input_field_types = {
                str: "String",
                int: "Int",
                bool: "Boolean",
                float: "Float",
            }

            try:
                field_type = input_field_types[
                    rel_table.primary_key.columns.id.type.python_type
                ]
            except KeyError:
                raise KeyError(
                    "The value set as the primary key for the relationship is not valid"
                )

            input_required_field = input_field = field_type

            if "TOMANY" in direction:
                base_field = MagqlList(base_field)
                input_required_field = MagqlList(input_required_field)
                input_field = MagqlList(input_field)
            elif required:
                input_required_field = MagqlNonNull(input_required_field)

            if (
                field_name
                not in self.magql_types[self.magql_name + "InputRequired"].fields
            ):
                self.magql_types[self.magql_name + "InputRequired"].fields[
                    field_name
                ] = MagqlInputField(input_required_field)
            if field_name not in self.magql_types[self.magql_name + "Input"].fields:
                self.magql_types[self.magql_name + "Input"].fields[
                    field_name
                ] = MagqlInputField(input_field)
            if field_name not in self.magql_types[self.magql_name].fields:
                self.magql_types[self.magql_name].fields[field_name] = MagqlField(
                    base_field, None, Resolver()
                )
            if field_name not in self.magql_types[self.magql_name + "Filter"].fields:
                self.magql_types[self.magql_name + "Filter"].fields[
                    field_name
                ] = MagqlInputField(RelFilter)
        self.magql_types[self.magql_name + "Payload"] = MagqlNonNull(
            MagqlObjectType(
                self.magql_name + "Payload",
                {
                    "errors": MagqlField(MagqlList("String")),
                    "result": MagqlField(self.magql_name, None, ResultResolver()),
                },
            )
        )

        self.magql_types[self.magql_name + "ListPayload"] = MagqlNonNull(
            MagqlObjectType(
                self.magql_name + "ListPayload",
                {
                    "errors": MagqlField(MagqlList("String")),
                    "result": MagqlField(
                        MagqlList(self.magql_name), None, ResultResolver()
                    ),
                    "count": MagqlField("Int", None, CountResolver()),
                },
            )
        )



class ValidationFailedError(Exception):
    def __init__(self, errors):
        if not isinstance(errors, list):
            errors = [errors]
        self.errors = errors


class AuthorizationError(Exception):
    def __init__(self, errors):
        if not isinstance(errors, list):
            errors = [errors]
        self.errors = errors



import json

from .flask_magql_utils import place_files_in_operations


def parse_request(request):
    if request.mimetype == "multipart/form-data":
        operations = json.loads(request.form.get("operations", "{}"))
        files_map = json.loads(request.form.get("map", "{}"))

        # return {
        #     "query": form["query"],
        #     "variables": json.loads(form["variables"]),
        #     "files": request.files,
        # }

        return place_files_in_operations(operations, files_map, request.files)

    else:
        return request.get_json()



import pytest
from sqlalchemy import Column
from sqlalchemy import DECIMAL
from sqlalchemy.types import Float
from sqlalchemy.types import Integer
from sqlalchemy.types import Numeric
from sqlalchemy.types import REAL
from sqlalchemy.types import String
from sqlalchemy.types import VARCHAR

from magql.definitions import MagqlFloat
from magql.definitions import MagqlInt
from magql.definitions import MagqlString
from magql.filter import IntFilter
from magql.filter import StringFilter
from magql.type import get_magql_filter_type
from magql.type import get_magql_type


get_magql_type_parameters = [
    (Column(String), MagqlString),
    (Column(VARCHAR), MagqlString),
    (Column(Integer), MagqlInt),
    (Column(Float), MagqlFloat),
    (Column(Numeric), MagqlFloat),
    (Column(REAL), MagqlFloat),
    (Column(DECIMAL), MagqlFloat),
]


@pytest.mark.parametrize("types", get_magql_type_parameters)
def test_get_type(types):
    assert isinstance(get_magql_type(types[0]), types[1])


get_magql_filter_type_parameters = [
    (Column(String), None, StringFilter),
    (Column(VARCHAR), None, StringFilter),
    (Column(Integer), None, IntFilter),
]


@pytest.mark.parametrize("types", get_magql_filter_type_parameters)
def test_get_filter_type(types):
    assert get_magql_filter_type(types[0], types[1]) == types[2]



import pytest
from graphql import assert_valid_schema
from graphql import GraphQLArgument
from graphql import GraphQLBoolean
from graphql import GraphQLEnumType
from graphql import GraphQLField
from graphql import GraphQLFloat
from graphql import GraphQLID
from graphql import GraphQLInputField
from graphql import GraphQLInputObjectType
from graphql import GraphQLInt
from graphql import GraphQLList
from graphql import GraphQLNonNull
from graphql import GraphQLObjectType
from graphql import GraphQLString
from graphql import GraphQLUnionType
from graphql import validate_schema

from magql.convert import Convert
from magql.definitions import MagqlArgument
from magql.definitions import MagqlBoolean
from magql.definitions import MagqlEnumType
from magql.definitions import MagqlField
from magql.definitions import MagqlFloat
from magql.definitions import MagqlID
from magql.definitions import MagqlInputField
from magql.definitions import MagqlInputObjectType
from magql.definitions import MagqlInt
from magql.definitions import MagqlList
from magql.definitions import MagqlNonNull
from magql.definitions import MagqlObjectType
from magql.definitions import MagqlString
from magql.definitions import MagqlUnionType


class DummyInfo:  # noqa: E501
    def __init__(self, session):
        self.type_map = {
            "Test": GraphQLObjectType("Test", GraphQLField(GraphQLList(GraphQLString))),
            "TestEmptyObject": GraphQLObjectType("TestEmptyObject", {}),
            "TestNestedObjects": GraphQLObjectType(
                "TestNestedObjects",
                {
                    "EnumField": GraphQLField(
                        GraphQLEnumType("TestEnum", {"RED": 0, "GREEN": 1, "BLUE": 2})
                    ),
                    "InputField": GraphQLInputField(GraphQLNonNull(GraphQLInt)),
                    "List": GraphQLList(GraphQLString),
                    "InputObjectType": GraphQLInputObjectType(
                        "TestInputObject",
                        GraphQLNonNull(
                            GraphQLUnionType("TestUnion", [GraphQLString, GraphQLID])
                        ),
                    ),
                    "ArgumentType": GraphQLArgument(GraphQLBoolean),
                    "Float": GraphQLFloat,
                },
            ),
        }
        self.context = session


@pytest.fixture
def info(session):
    return DummyInfo(session)


@pytest.mark.parametrize(
    "to_convert",
    [
        MagqlObjectType("Test", MagqlField(MagqlList(MagqlString))),
        MagqlObjectType("TestEmptyObject", {}),
        MagqlObjectType(
            "TestNestedObjects",
            {
                "EnumField": MagqlField(
                    MagqlEnumType("TestEnum", {"RED": 0, "GREEN": 1, "BLUE": 2})
                ),
                "InputField": MagqlInputField(MagqlNonNull(MagqlInt)),
                "List": MagqlList(MagqlString),
                "InputObjectType": MagqlInputObjectType(
                    "TestInputObject",
                    MagqlNonNull(
                        MagqlUnionType(
                            "TestUnion", [MagqlString, MagqlID], print("test"),
                        )
                    ),
                ),
                "ArgumentType": MagqlArgument(MagqlBoolean),
                "Float": MagqlFloat(),
            },
        ),
    ],
)
def test_convert_object(info, session, to_convert):
    converted = to_convert.convert(info.type_map)
    assert isinstance(converted, GraphQLObjectType)
    assert converted in info.type_map.values()


def test_convert_manager(info, session, manager_collection):
    converted = Convert(manager_collection.manager_map.values())
    converted_schema = converted.generate_schema()
    schema_validator = validate_schema(converted_schema)
    assert schema_validator == []
    assert assert_valid_schema(converted_schema) is None



import logging

from inflection import underscore
from sqlalchemy import func
from sqlalchemy.orm import lazyload
from sqlalchemy.orm import subqueryload
from sqlalchemy_utils import ChoiceType
from sqlalchemy_utils import get_mapper

from .errors import AuthorizationError
from .errors import ValidationFailedError
from .filter import generate_filters
from .sort import generate_sorts


def js_underscore(word):
    # add config
    return underscore(word)


class Resolver:
    """
    The super class for all builtin CRUD magql resolvers. Establishes
    the call order of the resolve functions. Resolve is broken down into
    3 main sections, :func:`pre-resolve`, :func:`resolve`, and
    :func:`post-resolve`. :func:`re-resolve` allows modification of the
    arguments that will be passed to :func:`resolve`. :func:`resolve`
    is broken down into 4 sections, :func:`retrieve_value`, :func:`authorize`,
    :func:`validate`, and :func:`mutate`. :func:`resolve` returns a value
    that can be modified by :func:`post_resolve`, which can perform side
    effects such as commiting the session. The order of these functions
    is the suggested way of organizing the resolve order such that overriding
    and/or extending the functionality is easiest.
    """

    def __call__(self, parent, info, *args, **kwargs):
        """
        Default resolve method, establishes and executes the default
        resolve flow
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the resolved value that is returned to GraphQL
        """
        parent, info, args, kwargs = self.pre_resolve(parent, info, *args, **kwargs)
        try:
            resolved_value = self.resolve(parent, info, *args, **kwargs)
        except ValidationFailedError as validation_errors:
            return {"errors": list(validation_errors.args)}
        except AuthorizationError as authorize_errors:
            return {"errors": list(authorize_errors.args)}
        post_resolved_value = self.post_resolve(
            resolved_value, parent, info, *args, **kwargs
        )
        return post_resolved_value

    def pre_resolve(self, parent, info, *args, **kwargs):
        """
        Allows for modification of the passed parameters
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the parameters that will be passed to :func:`resolve`
        """
        return parent, info, args, kwargs

    def post_resolve(self, resolved_value, parent, info, *args, **kwargs):
        """
        Allows for modification of the returned value to GraphQl and
        performing of side effects
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: The value to be returned to GraphQL
        """
        return resolved_value

    def authorize(self, instance, parent, info, *args, **kwargs):
        """
        Provides a space to perform authorization. Raise an AuthError
        to stop execution and have the resolve method return an error
        message based on the string error in the error object
        :param instance: The value returned by :func:`retrieve_value`
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        """
        return None

    def validate(self, instance, parent, info, *args, **kwargs):
        """
        Provides a space to perform validation. Raise an ValidationError
        to stop execution and have the resolve method return an error
        message based on the string error in the error object
        :param instance: The value returned by :func:`retrieve_value`
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        """
        return None

    def mutate(self, value, parent, info, *args, **kwargs):
        """
        Provides a space to mutate the resolved value. Necessarily will
        do nothing in queries.
        :param value: The value returned by
        :func:`retrieve_value`
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the value that will be returned to GraphQL
        """
        return value

    def retrieve_value(self, parent, info):
        """
        Retrieves (or creates) the value that the resolver will operate
        on and return. By default performs dot operation on the parent
        object using the field_name parameter of the info dict
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the value that will be operated on and returned to GraphQL
        """
        return getattr(parent, underscore(info.field_name))

    def resolve(self, parent, info, *args, **kwargs):
        """
        Establishes the call order of the resolve sub_functions
        :func:`retrieve_value`, :func:`authorize`, :func:`validate`,
        and :func:`mutate`. When subclassing, define one of these
        subfunctions to overwrite or extend the functionality in a
        granular manner. To override functionality in a more major way
        define a new resolve function to perform desired behavior.
        :param parent: gql parent. is whatever was returned by the
        parent resolver
        :param info: gql info dictionary
        :return: The value to be returned to GraphQL
        """
        value = self.retrieve_value(parent, info, *args, **kwargs)

        self.authorize(value, parent, info, *args, **kwargs)

        self.validate(value, parent, info, *args, **kwargs)

        value = self.mutate(value, parent, info, *args, **kwargs)

        return value


class ResultResolver:
    """
    Result Resolver retrieves the result key off of the payload object
    :param parent: GraphQL Payload object
    :param info: gql info dictionary
    :return: The results field on the payload object
    """

    def __call__(self, parent, info, *args, **kwargs):
        return parent


class CountResolver:
    def __call__(self, parent, info, *args, **kwargs):
        return info.context.info.get("count")


class CamelResolver:
    """
    Identical to graphql's default_field_resolver except the field_name
    is converted to snake case
    :param parent: gql parent. is whatever was returned by the
    parent resolver
    :param info: gql info dictionary
    :return: The value to be returned to GraphQL
    """

    def __call__(self, parent, info, *args, **kwargs):
        source = parent
        # TODO: Look into a way to generate info
        #  dictionary so the code does not need to be
        # copied or circumvent all together in a different way
        field_name = underscore(info.field_name)
        value = (
            source.get(field_name)
            if isinstance(source, dict)
            else getattr(source, field_name, None)
        )
        if callable(value):
            return value(info, **args)
        return value


class CheckDeleteResolver:
    """
    Resolver for the function that checks to see what will be deleted
    """

    def __init__(self, tables):
        self.tables = tables

    def __call__(self, parent, info, *args, **kwargs):
        for table in self.tables:
            try:
                class_ = get_mapper(table).class_
            except ValueError:
                continue
            # TODO: switch over frontend to class name
            if class_.__name__ == kwargs["tableName"]:
                id_ = kwargs["id"]
                session = info.context
                instance = session.query(class_).filter_by(id=id_).one()
                session.delete(instance)
                cascades = []
                for obj in session.deleted:
                    cascades.append(obj)

                session.rollback()

                return cascades


class SQLAlchemyTableUnionResolver:
    """
    Resolver that determines which type is being return from the delete check.
    This resolver is tied to the use of sqlalchemy
    """

    def __init__(self, magql_name_to_table):
        self.magql_name_to_table = magql_name_to_table

    def __call__(self, parent, info, *args, **kwargs):
        for magql_name, table in self.magql_name_to_table.items():
            if isinstance(parent, get_mapper(table).class_):
                for gql_type in info.return_type.of_type.types:
                    if gql_type.name == magql_name:
                        return gql_type
        raise Exception("Type not found")


class EnumResolver(Resolver):
    def retrieve_value(self, parent, info):
        """
        Resolve Enums which need to get the code from the value
        :param parent: gql parent. is whatever was returned by
        the parent resolver
        :param info: gql info dictionary
        :return: getattr(parent, info.field_Name)
        """
        if not parent:
            return None
        field_name = underscore(info.field_name)
        return getattr(getattr(parent, field_name), "code", None)


class TableResolver(Resolver):  # noqa: B903
    """
    A subclass of :class:`Resolver` that adds a table so that it can be
    reused in :class:`QueryResolver` and :class:`MutationResolver`.
    """

    def __init__(self, table):
        """
        MutationResolver can be overriden by
        :param table: a sqlalchemy table
        """
        self.table = table
        self.table_class = get_mapper(table).class_
        super().__init__()


class MutationResolver(TableResolver):
    """
    Subclass of :class:`TableResolver`. Initialized with a schema for
    validating the inputs. Its resolve method will call its validate
    method and return either the requested data of the changed object
    or an error dict filled with errors.
    """

    # def __call__(self, parent, info, *args, **kwargs):
    #     """
    #     Checks if a schema has been set and if so validates the mutation.
    #     Otherwise it just resolves the mutation.
    #     :param parent: parent object required by GraphQL, always None because
    #     mutations are always top level.
    #     :param info: GraphQL info dictionary
    #     :param args: Not used in automatic generation but left in in case
    #     overriding the validate or call methods.
    #     :param kwargs: Holds user inputs.
    #     :return:
    #     """
    #     return super(MutationResolver, self).__call__(parent, info, *args, **kwargs)

    def input_to_instance_values(self, input, mapper, session):
        """
        Helper method that converts the values in the input into values
        that can be passed into the creation of the instance. This
        returns scalars as themselves and passed id's as an object or
        list of objects that can be set in the creation call.
        :param input: The user's input dictionary containing the fields
        that are desired to be created.
        :param mapper: The mapper for the table that is being created
        :param session: The SQLAlchemy session
        :return: A dict of field names to values that will be added/changed on
        the newly created/modified object.
        """
        instance_values = {}
        for key, value in input.items():
            key = underscore(key)
            if key in mapper.c:
                col_type = mapper.c[key].type
                if isinstance(col_type, ChoiceType):
                    for enum_tuple in col_type.choices:
                        if value == enum_tuple[1]:
                            value = enum_tuple[0]
                            break
            if key in mapper.relationships:
                target = get_mapper(mapper.relationships[key].target).class_
                query = session.query(target)
                if value:
                    if isinstance(value, list):
                        value = query.filter(target.id.in_(value)).all()
                    else:
                        value = query.filter(target.id == value).one()
            instance_values[key] = value
        return instance_values

    # Post resolve will add and commit the created value
    def post_resolve(self, resolved_value, parent, info, *args, **kwargs):
        """
        Adds and commits the mutated value to the session
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: The value to be returned to GraphQL
        """
        session = info.context
        session.add(resolved_value)
        session.commit()
        return resolved_value


class ModelInputResolver(MutationResolver):
    def pre_resolve(self, parent, info, *args, **kwargs):
        """
        Converts ids of rels to actual values and handles enums
        :param parent: parent object required by GraphQL, always None because
        mutations are always top level.
        :param info: GraphQL info dictionary
        :param args: Not used in automatic generation but left in in case
        overriding the validate or call methods.
        :param kwargs: Holds user inputs.
        :return: The modified arguments
        """
        session = info.context
        mapper = get_mapper(self.table)
        kwargs["input"] = self.input_to_instance_values(
            kwargs["input"], mapper, session
        )

        return parent, info, args, kwargs


class CreateResolver(ModelInputResolver):
    """
    A subclass of :class:`MutationResolver`. Takes a dict of field values
    as input and creates an instance of the associated table with those
    fields.
    """

    def retrieve_value(self, parent, info, *args, **kwargs):
        """
        Creates an empty row in the table that will be modified by mutate.
        :param parent: parent object required by GraphQL, always None because
        mutations are always top level.
        :param info: GraphQL info dictionary
        :param args: Not used in automatic generation but left in in case
        overriding the validate or call methods.
        :param kwargs: Holds user inputs.
        :return: The instance with newly modified values
        """
        mapper = get_mapper(self.table)

        # TODO: Replace with dictionary spread operator
        instance = mapper.class_()
        return instance

    def mutate(self, instance, parent, info, *args, **kwargs):
        """
        Updates the passed instance to have the values specified
        in the query arguments
        :param value: The value returned by
        :func:`retrieve_value`
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the newly created value that will be returned to GraphQL
        """
        for key, value in kwargs["input"].items():
            setattr(instance, key, value)
        return instance


class UpdateResolver(ModelInputResolver):
    """
    A subclass of :class:`MutationResolver`. Takes a dict of field values
    and an id as input and updates the instance specified by id with
    fields specified by fields.
    """

    def retrieve_value(self, parent, info, *args, **kwargs):
        """
        Updates the instance of the associated table with the id passed.
        Performs setattr on the key/value pairs.
        :param parent: parent object required by GraphQL, always None because
        mutations are always top level.
        :param info: GraphQL info dictionary
        :param args: Not used in automatic generation but left in in case
        overriding the validate or call methods.
        :param kwargs: Holds user inputs.
        :return: The instance with newly modified valuesf
        """
        session = info.context
        mapper = get_mapper(self.table)

        id_ = kwargs["id"]
        return session.query(mapper.class_).filter_by(id=id_).one()

    def mutate(self, instance, parent, info, *args, **kwargs):
        """
        Updates the passed instance to have the values specified
        in the query arguments
        :param value: The value returned by
        :func:`retrieve_value`
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the updated value that will be returned to GraphQL
        """
        for key, value in kwargs["input"].items():
            setattr(instance, key, value)
        return instance


class DeleteResolver(MutationResolver):
    """
    A subclass of :class:`MutationResolver`. Takes an id and deletes
    the instance specified by id.
    """

    def retrieve_value(self, parent, info, *args, **kwargs):
        """
        Retrieves the row in the table that matches the id in the args,
        if such a row exists
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the value that will be operated on and returned to GraphQL,
        in this case the row with id matching the requested id
        """
        session = info.context
        mapper = get_mapper(self.table)
        id_ = kwargs["id"]
        return session.query(mapper.class_).filter_by(id=id_).one()

    def mutate(self, instance, parent, info, *args, **kwargs):
        """
        Deletes the passed instance from the session
        :param value: The value returned by
        :func:`retrieve_value`
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the value that was deleted
        """
        return instance

    def post_resolve(self, resolved_value, parent, info, *args, **kwargs):
        """
        Deletes the value from the session and commits the deletion
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: The value to be returned to GraphQL, in this case the
        deleted value
        """
        session = info.context
        session.delete(resolved_value)
        session.commit()
        return resolved_value


class QueryResolver(TableResolver):
    """
    A subclass of :class:`TableResolver`, the super class for
    :class:`SingleResolver` and :class:`ManyResolver`
    """

    def generate_query(self, info):
        """
        Generates a basic query based on the mapped class
        :param info: GraphQL info dict, used to hold the SQLA session
        :return: A SQLAlchemy query based on the mapped class,
        session.query(ModelClass)
        """
        session = info.context
        mapper = get_mapper(self.table)
        return session.query(mapper.class_)


class SingleResolver(QueryResolver):
    """
    A subclass of :class:`QueryResolver`. Takes an id and queries for
    the instance specified by id.
    """

    def post_resolve(self, resolved_value, parent, info, *args, **kwargs):
        return resolved_value if resolved_value else {"result": resolved_value}

    def retrieve_value(self, parent, info, *args, **kwargs):
        """
        Retrieves the row in the table that matches the id in the args,
        if such a row exists
        :param parent: gql parent. The value returned by the
        parent resolver. See GraphQL docs for more info
        :param info: GraphQL info dictionary, see GraphQL docs for more
        info
        :return: the value that will be operated on and returned to GraphQL,
        in this case the row with id matching the requested id
        """
        query = self.generate_query(info)
        return query.filter_by(id=kwargs["id"]).one_or_none()


class ManyResolver(QueryResolver):
    """
    A subclass of :class:`QueryResolver`. By default queries for all
    instances of the table that it is associated with. Can be filtered and
    sorted with keyword args.
    """

    def get_count(self, q):
        count_func = func.count()
        count_q = (
            q.options(lazyload("*"))
            .statement.with_only_columns([count_func])
            .order_by(None)
        )
        return q.session.execute(count_q).scalar()

    def generate_subqueryloads(self, field_node, load_path=None):
        """
        A helper function that allows the generation of the top level
        query to only have to perform one query with subqueryloads to
        eager load the data that will be accessed due to the structure
        of the query. Recursively builds a list of subquery loads
        that are applied to the base query.
        :param field_node: The document ast node that is used to determine
        what relationships are accessed by the query
        :param load_path: The load path that should be appended to
        in order to build the correct subquery
        :return:  A list of all subqueries needed to eagerly load
        all data accessed as a result of the query
        """
        options = []

        # A node is a lead if all of its children are scalars
        for selection in field_node.selection_set.selections:
            # selections with no sub selection_sets are scalars
            if selection.selection_set is None:
                continue
            field_name = js_underscore(selection.name.value)

            if field_name not in get_mapper(self.table).relationships:
                continue

            if load_path is None:
                extended_load_path = subqueryload(field_name)
            else:
                extended_load_path = load_path.subqueryload(field_name)
            options = options + self.generate_subqueryloads(
                selection, extended_load_path
            )

        # if all children are leaves then this is the last node,
        if len(options) == 0:
            return [load_path] if load_path is not None else []
        return options

    def generate_query(self, info):
        """
        Generates a query based on the document ast
        :param info: GraphQL info dict.
        :return: A SQLAlchemy query with any needed subquery loads
        appended
        """
        field_name = info.field_name
        field_node = [
            selection
            for selection in info.operation.selection_set.selections
            if selection.name.value == field_name
        ]
        if len(field_node) != 1:
            logging.getLogger(__name__).warning(
                f"Duplicate queries defined for {field_name!r}."
            )
        options = self.generate_subqueryloads(field_node[0])
        query = QueryResolver.generate_query(self, info).distinct()
        for option in options:
            query = query.options(option)

        return query

    def retrieve_value(self, parent, info, *args, **kwargs):
        """
        Returns all rows in a table. Uses subquery loads to improve
        querying by loading each relationship based on the query request
        :param parent: parent object required by GraphQL, always None because
        mutations are always top level.
        :param info:
        :param args: Not used but left in so that it can be used if a
        method is overriden
        :param kwargs: Holds the filters and sorts that can be used
        :return: A list of all rows in the table that match the
        filter, sorted by the given sort parameter.
        """
        query = self.generate_query(info)

        filters = generate_filters(self.table, info, **kwargs)
        for filter_ in filters:
            query = query.filter(filter_)
        sorts = generate_sorts(self.table, info, **kwargs)
        for sort in sorts:
            query = query.order_by(sort)

        paginated = False

        if info.variable_values.get("page") is not None:
            paginated = True
            current = info.variable_values.get("page").get("current")
            per_page = info.variable_values.get("page").get("per_page")
            if current is None or current < 1:
                current = 1
            if per_page is None or per_page < 1:
                per_page = 10

        info.context.info["count"] = self.get_count(query)

        if paginated:
            offset = (current - 1) * per_page
            return query.distinct().limit(per_page).offset(offset).all()

        return query.all()



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="magql",
    install_requires=[
        "graphql-core>=3",
        "inflection>=0.3",
        "SQLAlchemy>=1.3",
        "SQLAlchemy-Utils>=0.35",
    ],
)



import nox


@nox.session(python=["3.8", "3.7", "3.6"])
def tests(session):
    session.install(".", "-r", "requirements/tests.txt")
    session.run("pytest", *session.posargs)


@nox.session()
def style(session):
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", "--show-diff-on-failure")


@nox.session()
def docs(session):
    session.install(".", "-r", "requirements/docs.txt")
    session.run("sphinx-build", "-M", "clean", "docs", "docs/_build")
    session.run("sphinx-build", "-M", "html", "docs", "docs/_build", "-W")



import pytest
from graphql import execute
from graphql import parse

from .conftest import Car
from magql.convert import Convert


query_pagination = """
query test($page: Page) {
  cars (page: $page){
    result {
      name
    }
    count
  }
}
"""


def generate_page(current, per_page):
    return {"page": {"current": current, "per_page": per_page}}


@pytest.fixture
def schema(manager_collection):
    manager_list = list(manager_collection.manager_map.values())
    return Convert(manager_list).generate_schema()


def test_single_query(session, schema, single_query):
    document = parse(single_query)
    result = execute(schema, document, context_value=session)
    assert result.data["person"]["result"]["name"] == "Person 1"
    assert result.data["car"]["result"]["name"] is None
    assert result.data["house"]["result"]["name"] is None


def test_many_query(session, schema, many_query):
    document = parse(many_query)
    result = execute(schema, document, context_value=session)
    assert result.data["people"]["result"][0]["name"] == "Person 1"
    assert result.data["cars"]["result"][0]["name"] == "Car 1"
    assert result.data["houses"]["result"][0]["name"] == "House 1"


@pytest.mark.parametrize(
    ("page", "expected_count", "result_length"),
    [
        (generate_page(1, 1), 3, 1),
        (generate_page(2, 1), 3, 1),
        (generate_page(1, 2), 3, 2),
        (generate_page(-1, 2), 3, 2),
        (generate_page(-1, 2), 3, 2),
        (generate_page(6, 1), 3, 0),
        (generate_page(5, 5), 3, 0),
        (generate_page(1, 100), 3, 3),
    ],
)
def test_page(session, schema, page, expected_count, result_length):
    document = parse(query_pagination)

    car2 = Car(name="Car 2")
    car3 = Car(name="Car 3")
    session.add(car2)
    session.add(car3)
    session.commit()

    result = execute(schema, document, context_value=session, variable_values=page)
    assert len(result.data["cars"]["result"]) == result_length
    assert result.data["cars"]["count"] == expected_count



from graphql import GraphQLBoolean
from graphql import GraphQLFloat
from graphql import GraphQLID
from graphql import GraphQLInt
from graphql import GraphQLObjectType
from graphql import GraphQLSchema
from graphql import GraphQLString

from .definitions import MagqlBoolean
from .definitions import MagqlEnumType
from .definitions import MagqlField
from .definitions import MagqlFile
from .definitions import MagqlFloat
from .definitions import MagqlID
from .definitions import MagqlInputField
from .definitions import MagqlInt
from .definitions import MagqlString
from .definitions import MagqlUnionType
from .definitions import MagqlWrappingType


class Convert:
    """
    The Convert class is responsible for taking in a finalized magql
    schema and turning it into a graphql schema. Takes in the magql
    schema in the form of a list of managers. The magql schema from
    each manager is extracted and built into one magql schema. That
    schema is then converted to a graphql schema once
    :func:`generate_schema` is called.
    """

    def __init__(self, manager_list):
        """
        Generates a graphql schema based on the passed list of managers
        :param manager_list: A list of managers from which the Magql
        schema is extracted and turned into a graphql schema
        """
        # type_map maps magql_names to GraphQL and is needed to translate
        # MagqlTypes to GraphQLTypes
        self.gql_queries = {}
        self.gql_mutations = {}

        self.type_map = {
            "String": GraphQLString,
            "Int": GraphQLInt,
            "Boolean": GraphQLBoolean,
            "Float": GraphQLFloat,
            "ID": GraphQLID,
        }

        self.generate_type_map(manager_list)

        for manager in manager_list:
            if manager:
                self.convert_manager(manager)

    # TODO: Update convert_str_leafs and convert_type to either recursive or
    # functions on MagqlTypes
    @staticmethod
    def convert_type(ret_type, magql_type_map):
        wrapping_types_stack2 = []
        while isinstance(ret_type, MagqlWrappingType):
            wrapping_types_stack2.append(type(ret_type))
            ret_type = ret_type.type_
        if isinstance(ret_type, str):
            ret_type = magql_type_map[ret_type]
        while wrapping_types_stack2:
            ret_type = wrapping_types_stack2.pop()(ret_type)
        return ret_type

    @staticmethod
    def convert_str_leafs(type_, magql_type_map):
        wrapping_types_stack = []
        while isinstance(type_, MagqlWrappingType):
            wrapping_types_stack.append(type(type_))
            type_ = type_.type_
        if isinstance(type_, MagqlEnumType):
            return

        if isinstance(type_, MagqlUnionType):
            return
        for field_name, field in type_.fields.items():
            if not (
                isinstance(field, MagqlField) or isinstance(field, MagqlInputField)
            ):
                raise Exception(
                    f"Expected type MagqlField, got type {type(field)} "
                    f"for field: {field_name}\n Did you "
                    "forget to wrap your field with MagqlField?"
                )

            field.type_name = Convert.convert_type(field.type_name, magql_type_map)
            if not isinstance(field, MagqlInputField):
                for arg_name, arg in field.args.items():
                    field.args[arg_name] = Convert.convert_type(
                        arg.type_, magql_type_map
                    )

    def generate_type_map(self, managers):
        """
        Creates a mapping of String representations of Magql Types to
        the Magql types. Then converts the magql_types to grapqhl types.
        :param managers: The list of managers used to generate the
        GraphQL schema
        """
        magql_type_map = {
            "String": MagqlString(),
            "Int": MagqlInt(MagqlInt.parse_value_accepts_string),
            "Boolean": MagqlBoolean(),
            "Float": MagqlFloat(MagqlFloat.parse_value_accepts_string),
            "File": MagqlFile(),
            "ID": MagqlID(),
        }
        # Gather all magql types in magql_type_map
        for manager in managers:
            if not manager:
                continue
            for type_name, type_ in manager.magql_types.items():
                magql_type_map[type_name] = type_
        # replace string type representations with references to correct type
        for manager in managers:
            if not manager:
                continue
            for _type_name, type_ in manager.magql_types.items():
                Convert.convert_str_leafs(type_, magql_type_map)
            Convert.convert_str_leafs(manager.query, magql_type_map)
            Convert.convert_str_leafs(manager.mutation, magql_type_map)

        # TODO: Look into whether this convert_types is necessary and,
        # if it is whether or not it should be here
        for manager in managers:
            if manager:
                self.convert_types(manager)

    def convert_types(self, manager):
        for _magql_name, magql_type in manager.magql_types.items():
            magql_type.convert(self.type_map)
        for _magql_name, magql_type in manager.magql_types.items():
            magql_type.convert(self.type_map)

    def generate_mutations(self, manager):
        for _mutation_name, magql_field in manager.mutation.fields.items():
            for _argument_name, argument in magql_field.args.items():
                argument.convert(self.type_map)

            if magql_field.type_name not in self.type_map:
                pass

    def convert_manager(self, manager):
        # TODO: Consider moving the conversion of the query into the
        #  manager
        for query_name, query in manager.query.fields.items():
            self.gql_queries[query_name] = query.convert(self.type_map)

        for mut_name, mutation in manager.mutation.fields.items():
            self.gql_mutations[mut_name] = mutation.convert(self.type_map)

    def generate_schema(self):
        """
        Finalizes the GraphQL schema by generating the GraphQLSchema
        :return: The finalized GraphQL schema
        """
        query = GraphQLObjectType("Query", self.gql_queries)
        mutation = GraphQLObjectType("Mutation", self.gql_mutations)
        return GraphQLSchema(query, mutation)



from functools import wraps

from graphql import GraphQLArgument
from graphql import GraphQLBoolean
from graphql import GraphQLEnumType
from graphql import GraphQLField
from graphql import GraphQLFloat
from graphql import GraphQLID
from graphql import GraphQLInputField
from graphql import GraphQLInputObjectType
from graphql import GraphQLInt
from graphql import GraphQLList
from graphql import GraphQLNonNull
from graphql import GraphQLObjectType
from graphql import GraphQLString
from graphql import GraphQLUnionType
from graphql.type.scalars import coerce_float
from graphql.type.scalars import coerce_int
from inflection import camelize

RESERVED = [
    "JSON",
    "JSONType",
    "DateTime",
    "Text",
    "Date",
    "UnicodeText",
    "Unicode",
    "UrlType",
    "PhoneNumberType",
    "EmailType",
    "Time",
    "String",
    "VARCHAR",
    "Float",
    "Numeric",
    "Boolean",
    "ChoiceType",
]


class NamingConflictError(Exception):
    def __init__(self, name, magql_type):
        super().__init__((name, magql_type))
        self.name = name
        self.magql_type = magql_type

    def __str__(self):
        return f"{self.magql_type} instance cannot use reserved name {self.name}"


def check_name(init):
    @wraps(init)
    def wrapper(*args):
        if args[1] in RESERVED:
            raise NamingConflictError(args[1], args[0].__class__.__name__)
        init(*args)

    return wrapper


class MagqlObjectType:
    @check_name
    def __init__(self, name, fields=None, description=None):
        self.name = name
        # dict of field_name to MagqlField
        self.fields = fields if fields is not None else {}
        self.description = description

    def field(self, field_name, return_type, args=None):
        def decorator(resolve):
            self.fields[field_name] = MagqlField(return_type, args, resolve)
            return resolve

        return decorator

    # Convert each value in fields to GQLField
    def convert(self, type_map):
        if self.name in type_map:
            return type_map[self.name]
        type_map[self.name] = GraphQLObjectType(
            self.name, {}, None, description=self.description
        )
        for field_name, field in self.fields.items():
            type_map[self.name].fields[field_name] = field.convert(type_map)
        return type_map[self.name]


class MagqlField:
    def __init__(
        self,
        type_name=None,
        args=None,
        resolve=None,
        description=None,
        deprecation_reason=None,
    ):
        self.description = description
        self.deprecation_reason = deprecation_reason

        # String name representing type
        self.type_name = type_name
        self.args = args if args is not None else {}

        self.resolve = resolve

    def convert(self, type_map):
        gql_args = {}
        for arg_name, arg in self.args.items():
            gql_args[arg_name] = arg.convert(type_map)
        if self.type_name in type_map:
            field_type = type_map[self.type_name]
        else:
            field_type = self.type_name.convert(type_map)
        return GraphQLField(field_type, gql_args, self.resolve)


def js_camelize(word):
    # add config check
    # disable while camelcasing resolvers aren't added
    return camelize(word, False)


class MagqlArgument:  # noqa: E501
    def __init__(self, type_, default_value=None):
        self.type_ = type_
        self.default_value = default_value

    def convert(self, type_map):
        if self.type_ in type_map:
            converted_type = type_map[self.type_]
        else:
            converted_type = self.type_.convert(type_map)
        return GraphQLArgument(converted_type, self.default_value)


class MagqlInputObjectType:
    @check_name
    def __init__(self, name, fields=None, description=None):
        self.name = name
        self.fields = fields if fields is not None else {}
        self.description = description

    def convert(self, type_map):
        if self.name in type_map:
            return type_map[self.name]

        type_map[self.name] = GraphQLInputObjectType(self.name, {}, self.description)

        for field_name, field in self.fields.items():
            type_map[self.name].fields[field_name] = field.convert(type_map)

        return type_map[self.name]


class MagqlInputField:
    def __init__(self, type_name, description=None):
        self.type_name = type_name
        self.description = description

    def convert(self, type_map):
        if self.type_name in type_map:
            field_type = type_map[self.type_name]
        else:
            field_type = self.type_name.convert(type_map)
        return GraphQLInputField(field_type)


class MagqlWrappingType:
    pass


class MagqlNonNull(MagqlWrappingType):  # noqa: E501
    def __init__(self, type_):
        self.type_ = type_

    def convert(self, type_map):
        if self.type_ in type_map:
            return GraphQLNonNull(type_map[self.type_])
        return GraphQLNonNull(self.type_.convert(type_map))


class MagqlList(MagqlWrappingType):  # noqa: E501
    def __init__(self, type_):
        self.type_ = type_

    def convert(self, type_map):
        if self.type_ in type_map:
            converted_type = type_map[self.type_]
        else:
            converted_type = self.type_.convert(type_map)
        return GraphQLList(converted_type)


class MagqlEnumType:
    @check_name
    def __init__(self, name, values=None):
        self.name = name
        self.values = values if values else {}

    def convert(self, type_map):
        if self.name in type_map:
            return type_map[self.name]
        type_map[self.name] = GraphQLEnumType(self.name, self.values)
        return type_map[self.name]


class MagqlUnionType:  # noqa: B903
    @check_name
    def __init__(self, name, types, resolve_type):
        self.name = name

        # List of magql_types or magql_names
        self.types = types

        self.resolve_types = resolve_type

    def convert(self, type_map):
        if self.name in type_map:
            return type_map[self.name]
        types = []

        for enum_type in self.types:
            if isinstance(enum_type, str):
                types.append(type_map[enum_type])
            else:
                types.append(enum_type)
        type_map[self.name] = GraphQLUnionType(self.name, types, self.resolve_types)
        return type_map[self.name]


class MagqlInt:
    def __init__(self, parse_value=None):
        self.parse_value = parse_value

    @staticmethod
    def parse_value_accepts_string(value):
        try:
            converted_value = int(value)
        except ValueError:
            converted_value = coerce_int(value)
        return converted_value

    def convert(self, type_map):
        gql_int = GraphQLInt
        if self.parse_value:
            gql_int.parse_value = self.parse_value
        return gql_int


class MagqlFloat:
    def __init__(self, parse_value=None):
        self.parse_value = parse_value

    @staticmethod
    def parse_value_accepts_string(value):
        try:
            converted_value = float(value)
        except ValueError:
            converted_value = coerce_float(value)
        return converted_value

    def convert(self, type_map):
        gql_float = GraphQLFloat
        if self.parse_value:
            gql_float.parse_value = self.parse_value
        return gql_float


class MagqlFile:
    pass


class MagqlBoolean:
    def convert(self, type_map):
        return GraphQLBoolean


class MagqlString:
    def convert(self, type_map):
        return GraphQLString


class MagqlID:
    def convert(self, type_map):
        return GraphQLID
